/*
 * MICROCHIP SOFTWARE NOTICE AND DISCLAIMER:  You may use this software, and any derivatives created by any person or
 * entity by or on your behalf, exclusively with Microchip?s products.  Microchip and its licensors retain all ownership
 * and intellectual property rights in the accompanying software and in all derivatives hereto.
 *
 * This software and any accompanying information is for suggestion only.  It does not modify Microchip?s standard
 * warranty for its products.  You agree that you are solely responsible for testing the software and determining its
 * suitability.  Microchip has no obligation to modify, test, certify, or support the software.
 *
 * THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * BUT NOT LIMITED TO, IMPLIED WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A PARTICULAR PURPOSE
 * APPLY TO THIS SOFTWARE, ITS INTERACTION WITH MICROCHIP?S PRODUCTS, COMBINATION WITH ANY OTHER PRODUCTS, OR USE IN
 * ANY APPLICATION.
 *
 * IN NO EVENT, WILL MICROCHIP BE LIABLE, WHETHER IN CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE OR BREACH OF
 * STATUTORY DUTY), STRICT LIABILITY, INDEMNITY, CONTRIBUTION, OR OTHERWISE, FOR ANY INDIRECT, SPECIAL, PUNITIVE,
 * EXEMPLARY, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, FOR COST OR EXPENSE OF ANY KIND WHATSOEVER RELATED TO THE
 * SOFTWARE, HOWSOEVER CAUSED, EVEN IF MICROCHIP HAS BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE FORESEEABLE.
 * TO THE FULLEST EXTENT ALLOWABLE BY LAW, MICROCHIP'S TOTAL LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO THIS
 * SOFTWARE WILL NOT EXCEED THE AMOUNT OF FEES, IF ANY, THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR THIS SOFTWARE.
 *
 * MICROCHIP PROVIDES THIS SOFTWARE CONDITIONALLY UPON YOUR ACCEPTANCE OF THESE TERMS.
 *
 *
 * File:        bm70.c
 * Date:        March 17, 2016
 * Compiler:    XC16 v1.26
 *
 * Functions to communicate with a BM70 Bluetooth LE module over a UART
 *
 */

#include <xc.h>
#include <string.h>
#include "definitions.h"
#include "uart.h"
#include "timers.h"
#include "bm70.h"
#include "bm70_configure.h"

//Static functions only for use in this file
int8_t bmLocalSendCommandAndWaitResponse(enum CommandOpcode, uint8_t, uint8_t *);
static void bmLocalWriteCommandToBuffer(struct MessageOut *);

//Static variables  only for use in this file
static struct MessageIn messageInBuffer[SIZE_MSG_IN_BUFFER];                    //Buffer for incoming messages from module
static uint8_t msgInRdIndex, msgInWrIndex;                                      //Indexes to track additions and removals from message buffer
static struct MessageOut messageOutBuffer[SIZE_MSG_OUT_BUFFER];                 //Buffer for outgoing messages to module
static uint8_t msgOutRdIndex, msgOutWrIndex;                                    //Indexes to track additions and removals from message buffer

//**********************************************************************************************************************
// Set up BM70 module
// Read EEPROM, compare with correct settings, write EEPROM if necessary

int8_t bmSetupModule(uint8_t *name, uint8_t *address)
{
    struct MessageIn messageIncoming;                                           //Structure to hold incoming message
    uint8_t i, correctVersionResponse[5] = {0, 1, 6, 1, 1};                     //Expected response to reading the version number 
    uint8_t alternateVersionResponse[5] = {0, 1, 4, 1, 1};                      //Expected response to reading the version number 
    
    msgInRdIndex = msgInWrIndex = 0;                                            //Initialize the incoming message buffer indices for later use in the bmIncomingTasks() function
    msgOutRdIndex = msgOutWrIndex = 0;                                          //Initialize the outgoing message buffer indices for later use in the bmOutgoingTasks() function

    if (cfConfigureModule() == FALSE) {                                         //Check that module is configured correctly and configure if not
        return FALSE;
    }
    uartFlushRxBuffer();                                                        //Clear any junk out the buffer
    if (bmLocalSendCommandAndWaitResponse(ReadLocalInformation, 0, 0) == FALSE) { //Send command to read firmware version and MAC address
        SYS_DEBUG_BreakPoint();
        return FALSE;
    }
    if (bmGetIncomingMessage(&messageIncoming) == TRUE) {                       //See if we have a message from the module
        if (messageIncoming.event == CommandComplete && messageIncoming.command == ReadLocalInformation) { //See if the message is the Command Complete response to the Read Local Information command
            for (i = 0; i < 5; i++) {
                if (messageIncoming.data[i] != correctVersionResponse[i]) {     //See if the module has the correct firmware version
                    if (messageIncoming.data[i] != alternateVersionResponse[i]) { //See if the module has the correct firmware version
                        SYS_DEBUG_BreakPoint();
                        return FALSE;
                    }
                }
            }
            for (i = 10; i > 4; i--) {                                          //Reverse the order of MAC address bytes
                *address++ = messageIncoming.data[i];                           //Save the MAC address
            }
        } //if (messageIncoming.event == CommandComplete &&
        else {                                                                  //Did not get the correct response to read local information command
            SYS_DEBUG_BreakPoint();
            return FALSE;
        }
    } //if (bmGetIncomingMessage(&messageIncoming) == TRUE)
    else {                                                                      //Did not get a response to read local information command
        SYS_DEBUG_BreakPoint();
        return FALSE;
    }
    
    timersWait1ms(1);                                                           //Wait
    if (bmLocalSendCommandAndWaitResponse(ReadDeviceName, 0, 0) == FALSE) {     //Send command to read the device name
        SYS_DEBUG_BreakPoint();
        uartIsDataAvailable();
        return FALSE;
    }
    if (bmGetIncomingMessage(&messageIncoming) == TRUE) {                       //See if we have a message from the module
        if (messageIncoming.event == CommandComplete &&                         //See if the message is the Command Complete response to the Read Device Name command
                messageIncoming.command == ReadDeviceName && messageIncoming.data[0] == 0) { //and returned Command Succeeded status
            for (i = 0; i < 7; i++) {                                           //Get the first six characters of the device name
                if (messageIncoming.data[i + 1] > 31) {                         //Check that the character is printable
                    name[i] = messageIncoming.data[i + 1];                      //Save the name character
                }
                else {
                    name[i] = 32;                                               //or use space character instead
                }
            }
        } //if (messageIncoming.event == CommandComplete &&
        else {                                                                  //Did not get the correct response to read device name command
            SYS_DEBUG_BreakPoint();
            return FALSE;
        }
    } //if (bmGetIncomingMessage(&messageIncoming) == TRUE)
    else {                                                                      //Did not get a response to read device name command
        SYS_DEBUG_BreakPoint();
        return FALSE;
    }

    return TRUE;                                                                //Successfully updated the settings in the radio
}

//**********************************************************************************************************************
//Only for initialization because code is blocking for 20ms. Sends a command and waits for a response to be received

int8_t bmLocalSendCommandAndWaitResponse(enum CommandOpcode command, uint8_t parameter, uint8_t *data)
{
    int8_t timerSend = 0;

    bmSendCommand(command, parameter, data);                                    //Send command to read firmware version and MAC address
    timerSend = timersGetAndStartTimer(50);                                     //Get a timer to wait 50ms
    while (timersIsTimerDone(timerSend) == FALSE) {                             //Loop for 50ms while sending command and receiving a response
        bmOutgoingTasks();                                                      //Handle outgoing packets
        if (bmIncomingTasks() == TRUE) {                                        //Handle incoming packets and see if we have a message from the module
            timersFreeTimer(timerSend);
            return TRUE;
        }
    }
    timersFreeTimer(timerSend);
    return FALSE;
}

//**********************************************************************************************************************
// Receive and parse incoming messages from BM7x module - return TRUE if new message decoded

int8_t bmIncomingTasks(void)
{
    static enum StateIncomingPacket {FrameError, StartFrame, LengthHigh, LengthLow, EventOpcode, CommandOpcode, DataPayload, Checksum} stateIncomingPacket = StartFrame;
    static struct MessageIn *messageIncoming = &messageInBuffer[0];             //Pointer to the current message being written into the message buffer
    static uint8_t *ptrData = 0;                                                //Pointer to data payload from incoming packet (initialized to avoid compiler warning)
    static uint8_t checksum = 0;                                                //Checksum (initialized to avoid compiler warning)
    static uint16_t length = 0;                                                 //Data length in incoming packet (initialized to avoid compiler warning)
    uint8_t input;                                                              //Byte received from module

    if (uartIsDataAvailable()) {                                                //Check for new byte from BT module
        input = uartReadRxBuffer();                                             //Read the byte
        checksum += input;                                                      //Sum new byte for checksum

        switch (stateIncomingPacket) {                                          //Check which stage of packet reception we are in
            case FrameError:                                                    //Had a frame error (no start, invalid length, bad checksum)
                stateIncomingPacket = StartFrame;                               //Now look for start of frame
                //TODO should add delay to flush out potential bad data packet that might contain a start byte
                //For now just fall through to look for a new start of frame
                //break;
                
            case StartFrame:                                                    //Are waiting to receive a start byte (0xaa)
                if (input == 0xaa) {                                            //Look for start byte
                    ptrData = &(messageIncoming->data[0]);                      //Initialize the pointer to the buffer for incoming data in the selected message
                    checksum = 0;                                               //Start new checksum calculation
                    stateIncomingPacket = LengthHigh;                           //Now look for high byte of packet length
                }
                break;

            case LengthHigh:                                                    //Receiving high byte of two byte length field
                length = input * 256;                                           //Save high byte of length
                stateIncomingPacket = LengthLow;                                //Now look for low byte of packet length
                break;

            case LengthLow:                                                     //Receiving low byte of two byte length field
                length += input;                                                //Save low byte of length
                if (length > 0 && length <= (SIZE_INCOMING_DATA + 1)) {         //Must be at least one byte for opcode but not more than buffer size
                    length--;                                                   //Reduce length by one to get data length without event opcode
                    stateIncomingPacket = EventOpcode;                          //Now look for event opcode
                }
                else {
                    SYS_DEBUG_BreakPoint();
                    stateIncomingPacket = FrameError;                           //Error because length is out of range
                }
                break;

            case EventOpcode:                                                   //Receiving event opcode
                messageIncoming->event = input;                                 //Save the event opcode
                if (input == NoEventOpcode) {                                   //See if the event was a response to a command
                    SYS_DEBUG_BreakPoint();
                }
                if (input == CommandComplete) {                                 //See if the event was a response to a command
                    if (length > 0) {                                           //See if there is another byte for the command opcode
                        length--;                                               //Reduce length by one to get data length without command opcode
                        stateIncomingPacket = CommandOpcode;                    //Now look for command opcode
                    }
                    else {
                        SYS_DEBUG_BreakPoint();
                        stateIncomingPacket = FrameError;                       //Error because length does not allow for command opcode
                    }
                }
                else {
                    if (length > 0) {                                           //See if there is data following the opcode
                        stateIncomingPacket = DataPayload;                      //Now look for data
                    }
                    else {
                        stateIncomingPacket = Checksum;                         //Now look for checksum because there is no data
                    }
                }
                messageIncoming->length = length;                               //Save data length, not counting byte(s) for opcode(s)
                break;

            case CommandOpcode:                                                 //Receiving command opcode
                messageIncoming->command = input;                               //In that case save the command
                if (length > 0) {                                               //See if there is data following the opcode
                    stateIncomingPacket = DataPayload;                          //Now look for data
                }
                else {
                    stateIncomingPacket = Checksum;                             //Now look for checksum because there is no data
                }
                break;

            case DataPayload:                                                   //Receiving data payload
                *ptrData++ = input;                                             //Save received byte as data
                if (--length == 0) {                                            //Count down the bytes being received and see if done 
                    stateIncomingPacket = Checksum;
                }
                break;

            case Checksum:                                                      //Receiving checksum
                if (checksum == 0) {                                            //Bytes should sum to multiple of 256
                    msgInWrIndex++;                                             //Fully received a new message, so increment message pointer to next message
                    if (msgInWrIndex >= SIZE_MSG_IN_BUFFER) {                   //Check if end of buffer
                        msgInWrIndex = 0;                                       //Wrap index to beginning
                    }
                    messageIncoming = &messageInBuffer[msgInWrIndex];           //Set pointer for next message
                    stateIncomingPacket = StartFrame;                           //Now wait for next start of frame
                    if (msgInWrIndex == msgInRdIndex) {                         //Check if we have overrun the buffer space
                        SYS_DEBUG_BreakPoint();
                        return ERROR;                                           //Return error
                    }
                    return TRUE;                                                //Return TRUE because we have received a new message
                }
                else {
                    SYS_DEBUG_BreakPoint();
                    stateIncomingPacket = FrameError;                           //Error because of incorrect checksum
                }
                break;

            default:
                SYS_DEBUG_BreakPoint();
                stateIncomingPacket = FrameError;                               //Error because of invalid state
        }
    }

    return FALSE;                                                               //Did not get a complete packet
}
    
//**********************************************************************************************************************
// Get a message from the message buffer if available, return FALSE if no message in message buffer

int8_t bmGetIncomingMessage(struct MessageIn *message)
{
    if(msgInRdIndex != msgInWrIndex) {                                          //Check if message buffer is not empty
        *message = messageInBuffer[msgInRdIndex++];                             //Copy structure to return by reference and increment index (be careful, not all compilers will copy the content of the structure)
        if (msgInRdIndex >= SIZE_MSG_IN_BUFFER) {                               //Check if at end of buffer
            msgInRdIndex = 0;                                                   //then wrap the index to beginning
        }
        return TRUE;                                                            //Got a message so return TRUE 
    }
    return FALSE;                                                               //Buffer empty so return FALSE     
}

//**********************************************************************************************************************
// Send a command to the BM7x module

int8_t bmSendCommand(enum CommandOpcode command, uint8_t parameter, uint8_t *data)
{
    const struct MessageOut commandReadLocalInfo =              {.length = 1, .command = 0x01}; //all commands have length h/l, opcode, and optional data
    const struct MessageOut commandReset =                      {.length = 1, .command = 0x02};
    const struct MessageOut commandReadStatus =                 {.length = 1, .command = 0x03};
    const struct MessageOut commandReadADC =                    {.length = 2, .command = 0x04, .data = {0x00}};
    const struct MessageOut commandIntoShutdownMode =           {.length = 1, .command = 0x05};
    const struct MessageOut commandReadDeviceName =             {.length = 1, .command = 0x07};
    const struct MessageOut commandEraseAllPaired =             {.length = 1, .command = 0x09};
    const struct MessageOut commandReadPairingMode =            {.length = 1, .command = 0x0a};

    const struct MessageOut commandLeaveStandbyMode =           {.length = 2, .command = 0x1c, .data = {0x00}};
    const struct MessageOut commandEnterStandbyMode =           {.length = 2, .command = 0x1c, .data = {0x01}};
    const struct MessageOut commandEnterStandbyModeTrusted =    {.length = 2, .command = 0x1c, .data = {0x02}};
    const struct MessageOut commandEnterStandbyModeBeacon =     {.length = 2, .command = 0x1c, .data = {0x81}};
    const struct MessageOut commandEnterStandbyModeBeaconTrusted =  {.length = 2, .command = 0x1c, .data = {0x82}};

    const struct MessageOut commandReadAllPrimaryService =          {.length = 1, .command = 0x3b};
    const struct MessageOut commandReadLocalSpecificPrimaryService ={.length = 17, .command = 0x3c, .data = {0x8e, 0x95, 0x98, 0xac, 0x06, 0x62, 0x11, 0xe6, 0xb5, 0x12, 0x3e, 0x1d, 0x05, 0xde, 0xfe, 0x78}};

    struct MessageOut commandPacket;                                            //Structure to hold command packet
    uint8_t *commandDataPtr;                                                    //Pointer to access the data in the packet
    
    switch (command) {                                                          //Check which command is being sent
        case NoCommandOpcode:                                                   //No command selected
            return FALSE;
            break;
            
        //General commands
        case ReadLocalInformation:                                              //Send command to read firmware version and MAC address
            bmLocalWriteCommandToBuffer((struct MessageOut *)&commandReadLocalInfo);
            break;
            
        case Reset:                                                             //Send command to reset the module
            bmLocalWriteCommandToBuffer((struct MessageOut *)&commandReset);
            break;
            
        case ReadStatus:                                                        //Send command to read module status
            bmLocalWriteCommandToBuffer((struct MessageOut *)&commandReadStatus);
            break;
            
        case ReadADCValue:                                                      //Send command to read ADC (TODO add code for ADC channel)
            bmLocalWriteCommandToBuffer((struct MessageOut *)&commandReadADC);
            break;
            
        case IntoShutdownMode:                                                  //Send command to enter shutdown mode
            bmLocalWriteCommandToBuffer((struct MessageOut *)&commandIntoShutdownMode);
            break;

        case DebugCommand:
            break;

        case ReadDeviceName:                                                    //Send command to read device name
            bmLocalWriteCommandToBuffer((struct MessageOut *)&commandReadDeviceName);
            break;
            
        case WriteDeviceName:
            if (parameter == 0 || parameter > 20) {                             //Limit name from 1 to 20 characters
                return FALSE;                                                   //Invalid length so return fail
            }
            commandPacket.command = 0x08;                                       //Load the opcode for WriteDeviceName command
            commandPacket.length = parameter + 2;                               //Length must also include opcode, reserved byte
            commandDataPtr = &commandPacket.data[0];                            //Point to the data
            *commandDataPtr++ = 0;                                              //Copy reserved byte to data payload 
            while (parameter-- > 0) {
                *commandDataPtr++ = *data++;                                    //Copy name to data payload 
            }
            bmLocalWriteCommandToBuffer(&commandPacket);                        //Send the packet
            break;
            
        case EraseAllPairedDeviceInformation:                                   //Send command to erase all paired device information
            bmLocalWriteCommandToBuffer((struct MessageOut *)&commandEraseAllPaired);
            break;
            
        case ReadPairingModeSetting:                                            //Send command to read pairing mode setting
            bmLocalWriteCommandToBuffer((struct MessageOut *)&commandReadPairingMode);
            break;
            
        case WritePairingModeSetting:
        case ReadAllPairedDeviceInformation:
        case DeletePairedDevice:
        case DitalIOControl:
        case PWMControl:
        //GAP commands    
        case ReadRSSIValue:
        case WriteAdvertisingData:
        case WriteScanResultData:
        case SetAdvertisingParameter:
        case SetScanParameter:
        case SetScanEnable:
        case LeCreateConnection:
        case LeCreateConnectionCancel:
        case ConnectionParameterUpdateRequest:
        case Disconnect:
            return FALSE;
            break;

        case SetAdvertisingEnable:
            switch (parameter) {
                case LeaveStandbyMode:                                          //Leave standby mode
                    bmLocalWriteCommandToBuffer((struct MessageOut *)&commandLeaveStandbyMode); //Send command to leave standby mode
                    break;
                case EnterStandbyMode:                                          //Enter standby mode (starts advertising))
                    bmLocalWriteCommandToBuffer((struct MessageOut *)&commandEnterStandbyMode);
                    break;
                case EnterStandbyModeForTrusted:                                //Enter standby mode only connectable to trusted device
                    bmLocalWriteCommandToBuffer((struct MessageOut *)&commandEnterStandbyModeTrusted);
                    break;
                case EnterStandbyModeWithBeacon:                                //Enter standby mode with beacon enabled
                    bmLocalWriteCommandToBuffer((struct MessageOut *)&commandEnterStandbyModeBeacon);
                    break;
                case EnterStandbyModeWithBeaconForTrusted:                      //Enter standby mode  with beacon enabled and only connectable to trusted device
                    bmLocalWriteCommandToBuffer((struct MessageOut *)&commandEnterStandbyModeBeaconTrusted);
                    break;
                default:                                                        //Invalid setting
                    return ERROR;
            }
            break;
        case ReadRemoteDeviceName:
        //GATT Client commands    
        case DiscoverAllPrimaryServices:
        case DiscoverSpecificPrimaryServiceCharacteristics:
        case ReadCharacteristicValue:
        case ReadUsingCharacteristicUUID:
        case WriteCharacteristicValue:
        case EnableTransparent:
            break;
        //GATT Server commands    
        case SendCharacteristicValue:
            if (parameter == 0 || parameter > 20) {                             //Characteristic value must be 1 to 20 bytes
                return FALSE;                                                   //Invalid length so return fail
            }
            parameter += 4;                                                     
            commandPacket.command = 0x38;                                       //Load the opcode for SendCharacteristicValue command
            commandPacket.length = parameter;                                   //Length must also include opcode, connection handle and char value handle
            commandDataPtr = &commandPacket.data[0];                            //Point to the data
            while (parameter-- > 0) {
                *commandDataPtr++ = *data++;                                    //Copy connection handle, char value handle, and char value data 
            }
            bmLocalWriteCommandToBuffer(&commandPacket);                        //Send the packet
            break;
            
        case UpdateCharacteristicValue:
        case ReadLocalCharacteristicValue:
        case ReadLocalAllPrimaryService:                                        //Read all local services
            bmLocalWriteCommandToBuffer((struct MessageOut *)&commandReadAllPrimaryService);
            break;
        case ReadLocalSpecificPrimaryService:                                   //Read a specific (by UUIID) local service
            bmLocalWriteCommandToBuffer((struct MessageOut *)&commandReadLocalSpecificPrimaryService);
            break;
            
        //Transparent service command
        case SendTransparentData:
        //Pairing commands    
        case PasskeyEntryResponse:
        case UserConfirmResponse:
        case PairingRequest:
        //Configure mode command
        case LeaveConfigureMode:
        default:
            return FALSE;
    }
    return TRUE;
}


//**********************************************************************************************************************
// Copy a command to the output buffer so it can be sent to the module by bmOutgoingTasks()

static void bmLocalWriteCommandToBuffer(struct MessageOut *msg)
{
    messageOutBuffer[msgOutWrIndex++] = *msg;                                   //Copy structure and increment index (be careful, not all compilers will copy the content of the structure)
    if (msgOutWrIndex >= SIZE_MSG_OUT_BUFFER) {                                 //Check if end of buffer
        msgOutWrIndex = 0;                                                      //Wrap index to beginning
    }
    if (msgOutWrIndex == msgOutRdIndex) {                                       //Check if we have overrun the buffer space
        SYS_DEBUG_BreakPoint();
        //return ERROR;                                                         //Return error
    }
}


//**********************************************************************************************************************
// Send any messages queued in the send buffer

int8_t bmOutgoingTasks(void)
{
    static enum StateOutgoingPacket {OutgoingIdle, OutgoingRxIndLow, OutgoingTransmit, OutgoingRxIndHigh} stateOutgoingPacket = OutgoingIdle;
    static int8_t timerOutgoing = 0;
    uint16_t length;
    uint8_t *data, temp, checksum = 0;                                          //Initial value of checksum

    switch (stateOutgoingPacket) {
        case OutgoingIdle:                                                      //Idle, so keep checking for a new outgoing message
            if(msgOutRdIndex != msgOutWrIndex) {                                //Check if message buffer is not empty
                BLE_RX_IND_PIN = 0;                                             //Indicate that we are going to transmit so module can come out of power saving
                timerOutgoing = timersGetAndStartTimer(2);                      //Get a timer to wait 2ms
                stateOutgoingPacket = OutgoingRxIndLow;                         //Now wait for 2ms for module to wake up
            }
            break;
            
        case OutgoingRxIndLow:                                                  //Waiting for time since RX_IND pin went low so keep checking timer
            if (timersIsTimerDone(timerOutgoing)) {                             //Check whether we have waited for module to be ready to receive
                timersFreeTimer(timerOutgoing);                                 //Release the timer we used
                length = messageOutBuffer[msgOutRdIndex].length;
                if (length > 0 && length < SIZE_OUTGOING_DATA) {                //Check that length is within bounds
                    uartWriteTxBuffer(0xaa);                                    //Load start byte into the transmit buffer
                    temp = (uint8_t)(length >> 8);                              //Calculate high byte of length
                    uartWriteTxBuffer(temp);                                    //Load high byte of length into the transmit buffer
                    checksum -= temp;                                           //Do running checksum calculation
                    temp = (uint8_t)(length & 0x00ff);                          //Calculate low byte of length
                    uartWriteTxBuffer(temp);                                    //Load low byte of length into the transmit buffer
                    checksum -= temp;                                           //Do running checksum calculation
                    temp = messageOutBuffer[msgOutRdIndex].command;
                    uartWriteTxBuffer(temp);                                    //Load opcode into the transmit buffer
                    checksum -= temp;                                           //Do running checksum calculation
                    data = &messageOutBuffer[msgOutRdIndex].data[0];            //Load pointer to data
                    while (--length > 0) {
                        uartWriteTxBuffer(*data);                               //Load data into the transmit buffer
                        checksum -= *data++;                                    //Do running checksum calculation
                    }
                    uartWriteTxBuffer(checksum);                                //Load the checksum 
                    uartStartTransmitData();                                    //Start transmitting the bytes
                }
                if (msgOutRdIndex++ >= SIZE_MSG_OUT_BUFFER) {                   //Increment read index and check if at end of buffer
                    msgOutRdIndex = 0;                                          //then wrap the index to beginning
                }
                stateOutgoingPacket = OutgoingTransmit;
            }
            break;
            
        case OutgoingTransmit:
            if (uartIsBusyTransmitting() == FALSE) {                            //Wait for transmission to complete
                BLE_RX_IND_PIN = 1;                                             //Indicate done transmitting, module can go into power saving mode
                stateOutgoingPacket = OutgoingIdle;                             //Back to idle state
            }
            break;
            
        default:
            stateOutgoingPacket = OutgoingIdle;                                 //Back to idle state
            break;
    }
    
    return TRUE;
}

