/*
 * MICROCHIP SOFTWARE NOTICE AND DISCLAIMER:  You may use this software, and any derivatives created by any person or
 * entity by or on your behalf, exclusively with Microchip?s products.  Microchip and its licensors retain all ownership
 * and intellectual property rights in the accompanying software and in all derivatives hereto.
 *
 * This software and any accompanying information is for suggestion only.  It does not modify Microchip?s standard
 * warranty for its products.  You agree that you are solely responsible for testing the software and determining its
 * suitability.  Microchip has no obligation to modify, test, certify, or support the software.
 *
 * THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * BUT NOT LIMITED TO, IMPLIED WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A PARTICULAR PURPOSE
 * APPLY TO THIS SOFTWARE, ITS INTERACTION WITH MICROCHIP?S PRODUCTS, COMBINATION WITH ANY OTHER PRODUCTS, OR USE IN
 * ANY APPLICATION.
 *
 * IN NO EVENT, WILL MICROCHIP BE LIABLE, WHETHER IN CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE OR BREACH OF
 * STATUTORY DUTY), STRICT LIABILITY, INDEMNITY, CONTRIBUTION, OR OTHERWISE, FOR ANY INDIRECT, SPECIAL, PUNITIVE,
 * EXEMPLARY, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, FOR COST OR EXPENSE OF ANY KIND WHATSOEVER RELATED TO THE
 * SOFTWARE, HOWSOEVER CAUSED, EVEN IF MICROCHIP HAS BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE FORESEEABLE.
 * TO THE FULLEST EXTENT ALLOWABLE BY LAW, MICROCHIP'S TOTAL LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO THIS
 * SOFTWARE WILL NOT EXCEED THE AMOUNT OF FEES, IF ANY, THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR THIS SOFTWARE.
 *
 * MICROCHIP PROVIDES THIS SOFTWARE CONDITIONALLY UPON YOUR ACCEPTANCE OF THESE TERMS.
 *
 *
 * File:        bm70_configure.c
 * Date:        June 17, 2016
 * Compiler:    XC16 v1.26
 *
 * Functions to configure the BM70 Bluetooth LE module when it is used for the first time.
 * 
 * Code checks the configuration against the desired configuration and updates the configuration Flash memory
 * in the BM70 if necessary 
 *
 * The desired configuration is in a text file created using the BM7x EEPROM Table Utility (bm7x_eeprom_table_utility.exe)
 *
 */

#include <xc.h>
#include <string.h>
#include "definitions.h"
#include "uart.h"
#include "timers.h"
#include "bm70.h"

//Static functions only for use in this file
static void cfLocalEnterTestMode(void);
static void cfLocalEnterAppMode(void);
static int8_t cfLocalAccessFlash(void);
static int8_t cfLocalEraseFlash(void);
static int8_t cfLocalEndAccess(void);
static int8_t cfLocalVerifyConfiguration(void);
static int8_t cfLocalReadFlash(uint16_t, uint8_t *);
static int8_t cfLocalProgramConfiguration(void);
static int8_t cfLocalWriteFlash(uint16_t, uint8_t *);
static int8_t cfLocalSendCommandAndCheckResponse(uint8_t *, uint8_t, uint8_t *, uint8_t);

static const struct FlashConfigurationLine configFlashTable[] = {
#ifdef USE_BM70
    #include "bm70_config_lab4p_solution.txt"                                      //Use BM7x EEPROM Table Utility to create this file
#else
    #include "bm71_config_lab4p_solution.txt"                                       //Use BM7x EEPROM Table Utility to create this file
#endif
};

//**********************************************************************************************************************
// Configure the BM70 module
// Read configuration flash, compare with correct settings, erase and write if necessary

int8_t cfConfigureModule(void) {
    
    if (sizeof(configFlashTable) != 9728) {                                     //Check that our array of configuration flash values is the correct size 
        cfLocalEnterAppMode();                                                  //Reset the module and go into application mode
        SYS_DEBUG_BreakPoint();                                                 //Debug
        return FALSE;                                                           //Failed to access configuration flash
    }
    
    cfLocalEnterTestMode();                                                     //Reset the module and put into test mode to access configuration flash
    uartFlushRxBuffer();                                                        //Clear any junk out the buffer
    if (cfLocalAccessFlash() == FALSE) {                                        //Send command to access configuration flash and get response
        cfLocalEnterAppMode();                                                  //Reset the module and go into application mode
        SYS_DEBUG_BreakPoint();                                                 //Debug
        return FALSE;                                                           //Failed to access configuration flash
    }

    if(cfLocalVerifyConfiguration() == TRUE) {                                  //See if configuration of the module is correct
        cfLocalEnterAppMode();                                                  //Reset the module and go into application mode
        return TRUE;                                                            //Have correct configuration so return
    }

    if (cfLocalEraseFlash() == FALSE) {                                         //Send command to erase configuration flash and get response
        cfLocalEnterAppMode();                                                  //Reset the module and go into application mode
        SYS_DEBUG_BreakPoint();                                                 //Debug
        return FALSE;                                                           //Failed to erase configuration flash
    }
    //timersWait1ms(100);                                                       //Wait for configuration flash to be erased

    if(cfLocalProgramConfiguration() == FALSE) {                                //Program the configuration flash
        cfLocalEnterAppMode();                                                  //Reset the module and go into application mode
        SYS_DEBUG_BreakPoint();                                                 //Debug
        return TRUE;                                                            //Have correct configuration so return
    }

    if(cfLocalVerifyConfiguration() == FALSE) {                                 //See if configuration of the module is correct
        cfLocalEnterAppMode();                                                  //Reset the module and go into application mode
        SYS_DEBUG_BreakPoint();                                                 //Debug
        return FALSE;                                                           //Failed to erase configuration flash
    }

    if (cfLocalEndAccess() == FALSE) {                                          //Send command to end access to configuration flash and get response
        cfLocalEnterAppMode();                                                  //Reset the module and go into application mode
        SYS_DEBUG_BreakPoint();                                                 //Debug
        return FALSE;                                                           //Failed to end access
    }
    cfLocalEnterAppMode();                                                      //Reset the module and go into application mode
    return TRUE;
}

//**********************************************************************************************************************
// Reset module and put into test mode

static void cfLocalEnterTestMode(void)
{
    BLE_NOT_RESET_PIN = 0;                                                      //Reset the module
    timersWait1ms(1);                                                           //Hold BM70 reset pin low for at least 1ms
    BLE_MODE_PIN = 0;                                                           //Go into test mode to read or write configuration flash
    timersWait1ms(1);                                                           //Hold BM70 reset pin low for at least 1ms
    BLE_NOT_RESET_PIN = 1;                                                      //Release from reset
    timersWait1ms(100);                                                         //Wait at least 46ms for module to enter test mode
}

//**********************************************************************************************************************
// Reset module and put into application mode

static void cfLocalEnterAppMode(void)
{
    BLE_NOT_RESET_PIN = 0;                                                      //Reset the module
    timersWait1ms(1);                                                           //Hold BM70 reset pin low for at least 1ms
    BLE_MODE_PIN = 1;                                                           //Go into application mode
    timersWait1ms(1);                                                           //Hold BM70 reset pin low for at least 1ms
    BLE_NOT_RESET_PIN = 1;                                                      //Release from reset
    timersWait1ms(200);                                                         //Wait at least 68ms for module to enter application mode
}

//**********************************************************************************************************************
// Send command to access configuration flash and check response

int8_t cfLocalAccessFlash(void)
{
    uint8_t packetOut[17] = {0x01,0x05,0x04,0x0D,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}; //Access command packet
    uint8_t packetIn[21] = {0x04,0x0F,0x04,0x00,0x01,0x05,0x04,0x04,0x03,0x0B,0x00,0xFF,0x0F,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00};

    return cfLocalSendCommandAndCheckResponse(packetOut, 17, packetIn, 21);
}

//**********************************************************************************************************************
// Send command to erase configuration flash and check response

int8_t cfLocalEraseFlash(void)
{
    uint8_t packetOut[19] = {0x02,0xFF,0x0F,0x0E,0x00,0x12,0x01,0x0A,0x00,0x03,0x00,0x00,0x40,0x03,0x00,0x00,0x20,0x00,0x00}; //Erase command packet
    uint8_t packetIn[27] = {0x04,0x13,0x05,0x01,0xFF,0x0F,0x01,0x00,0x02,0xFF,0x0F,0x0E,0x00,0x12,0x01,0x0A,0x00,0x00,0x00,0x00,0x40,0x03,0x00,0x00,0x20,0x00,0x00};

    return cfLocalSendCommandAndCheckResponse(packetOut, 19, packetIn, 27);
}


//**********************************************************************************************************************
// Send command to end access to configuration flash and check response

int8_t cfLocalEndAccess(void)
{
    uint8_t packetOut[7] = {0x01,0x06,0x04,0x03,0xFF,0x0F,0x00};                //End access command packet
    uint8_t packetIn[14] = {0x04,0x0F,0x04,0x00,0x01,0x06,0x04,0x04,0x05,0x04,0x00,0xFF,0x0F,0x00};

    return cfLocalSendCommandAndCheckResponse(packetOut, 7, packetIn, 14);
}

//**********************************************************************************************************************
// Read the entire configuration flash and verify against the correct configuration

static int8_t cfLocalVerifyConfiguration(void)
{
    uint16_t i, j, k;
    uint16_t address;
    uint8_t dataIn[128], *ptrDataIn;                                            //Packets contain 128 bytes of data (after the 19 byte preamble))
    
    address = 0;
    for(i = 0; i < 512; i += 8) {
        if (cfLocalReadFlash(address, dataIn) == FALSE) {                       //Send command to read configuration flash and get response
            SYS_DEBUG_BreakPoint();                                             //Debug
            return FALSE;                                                       //Failed to read configuration flash
        }
        ptrDataIn = &dataIn[0];
        for (j = 0; j < 8; j++) {
            for (k = 0; k < 16; k++) {
                if(*ptrDataIn++ != configFlashTable[i+j].data[k]) {
                    return FALSE;
                }
            }
        }
        address += 128;
    }
    return TRUE;
}

//**********************************************************************************************************************
// Read 128 bytes of configuration flash

int8_t cfLocalReadFlash(uint16_t address, uint8_t *data)
{
    uint8_t i, packetOut[19] = {0x02,0xFF,0x0F,0x0E,0x00,0x10,0x01,0x0A,0x00,0x03,0x00,0x00,0x40,0x03,0x00,0x80,0x00,0x00,0x00}; //Read command packet
    uint8_t packetIn[19] = {0x04,0x13,0x05,0x01,0xFF,0x0F,0x01,0x00,0x02,0xFF,0x0F,0x86,0x00,0x10,0x01,0x82,0x00,0x00,0x00};
    int8_t timerSend = 0;
    
    packetOut[11] = (uint8_t)address;                                           //Add lower address byte to base address low byte (0x00) already in packet
    packetOut[12] += (uint8_t)(address >> 8);                                   //Add upper address byte to base address high byte (0x40) already in packet

    if (cfLocalSendCommandAndCheckResponse(packetOut, 19, packetIn, 19) == FALSE) {
        SYS_DEBUG_BreakPoint();                                                 //Debug
        return FALSE;                                                           //Did not get correct response
    }

    timerSend = timersGetAndStartTimer(50);                                     //Get a timer to wait 50ms for response
    for (i = 0; i < 128; i++) {                                                 //Should receive 128 bytes following the preamble
        while (uartIsDataAvailable() == FALSE) {                                //Wait for new byte from UART
            if (timersIsTimerDone(timerSend) == TRUE) {                         //See if 50ms time is up
                timersFreeTimer(timerSend);                                     //Free up the timer
                SYS_DEBUG_BreakPoint();                                         //Debug
                return FALSE;                                                   //Timed out while waiting for data
            }
        }
        *data++ = uartReadRxBuffer();                                           //Read the byte and put it in the return buffer
        if (timersIsTimerDone(timerSend) == TRUE) {                             //See if 50ms time is up
            timersFreeTimer(timerSend);                                         //Free up the timer
            SYS_DEBUG_BreakPoint();                                             //Debug
            return FALSE;                                                       //Timed out while waiting for data
        }
    }

    timersFreeTimer(timerSend);                                                 //Free up the timer
    return TRUE;                                                                //Got all 147 bytes
}

//**********************************************************************************************************************
// Write the entire configuration flash

int8_t cfLocalProgramConfiguration(void)
{
    uint16_t i, j, k;
    uint16_t address;
    uint8_t dataOut[128], *ptrDataOut;                                          //Packets contain 128 bytes of data (after the 19 byte preamble))
    
    address = 0;
    for(i = 0; i < 512; i += 8) {
        ptrDataOut = &dataOut[0];
        for (j = 0; j < 8; j++) {
            for (k = 0; k < 16; k++) {
                *ptrDataOut++ = configFlashTable[i+j].data[k];                  //Get a byte to program and add it to the data to write
            }
        }
        if (cfLocalWriteFlash(address, dataOut) == FALSE) {                     //Send command to write configuration flash and get response
            SYS_DEBUG_BreakPoint();                                             //Debug
            return FALSE;                                                       //Failed to read configuration flash
        }
        address += 128;
    }
    return TRUE;
}

//**********************************************************************************************************************
// Write 128 bytes of configuration flash

int8_t cfLocalWriteFlash(uint16_t address, uint8_t *data)
{
    uint8_t i, packetOut[147] = {0x02,0xFF,0x0F,0x8E,0x00,0x11,0x01,0x8A,0x00,0x03,0x00,0x00,0x40,0x03,0x00,0x80,0x00,0x00,0x00}; //Write command packet
    uint8_t packetIn[19] = {0x04,0x13,0x05,0x01,0xFF,0x0F,0x01,0x00,0x02,0xFF,0x0F,0x06,0x00,0x11,0x01,0x02,0x00,0x00,0x00};
    
    packetOut[11] = (uint8_t)address;                                           //Add lower address byte to base address low byte (0x00) already in packet
    packetOut[12] += (uint8_t)(address >> 8);                                   //Add upper address byte to base address high byte (0x40) already in packet

    for (i = 19; i < 147; i++) {                                                //Add data to outgoing packet with write command
        packetOut[i] = *data++;
    }
    
    if (cfLocalSendCommandAndCheckResponse(packetOut, 147, packetIn, 19) == FALSE) {
        SYS_DEBUG_BreakPoint();                                                 //Debug
        return FALSE;                                                           //Did not get correct response
    }
    return TRUE;
}

//**********************************************************************************************************************
// Send command and check response

int8_t cfLocalSendCommandAndCheckResponse(uint8_t *outgoing, uint8_t outgoingLength, uint8_t *incoming, uint8_t incomingLength)
{
    uint16_t i;
    int8_t timerSend = 0;
    
    for (i = 0; i < outgoingLength; i++) {
        uartWriteTxBuffer(*outgoing++);                                         //Load the packet into the transmit buffer 
    }
    uartStartTransmitData();                                                    //Start transmitting the bytes
  
    timerSend = timersGetAndStartTimer(50);                                     //Get a timer to wait 50ms for response
    for (i = 0; i < incomingLength; i++) {                                      //Loop through each byte of the response
        while (uartIsDataAvailable() == FALSE) {                                //Wait for new byte from UART
            if (timersIsTimerDone(timerSend) == TRUE) {                         //See if 50ms time is up
                timersFreeTimer(timerSend);                                     //Free up the timer
                SYS_DEBUG_BreakPoint();                                         //Debug
                return FALSE;                                                   //Timed out while waiting for data
            }
        }
        if (uartReadRxBuffer() != *incoming++) {                                //Read the byte and see if it matches what is expected
            while (timersIsTimerDone(timerSend) == FALSE) {                     //Wrong byte so wait and flush rest of packet before returning 
                if (uartIsDataAvailable()) {                                    //Check for new byte from BT module
                    uartReadRxBuffer();                                         //Read the byte and discard
                }
            }
            timersFreeTimer(timerSend);                                         //Free up the timer
            SYS_DEBUG_BreakPoint();                                             //Debug
            return FALSE;                                                       //Received bad data
        }
    }
    timersFreeTimer(timerSend);                                                 //Free up the timer
    return TRUE;                                                                //Got correct response
}

