/*
 * MICROCHIP SOFTWARE NOTICE AND DISCLAIMER:  You may use this software, and any derivatives created by any person or
 * entity by or on your behalf, exclusively with Microchip?s products.  Microchip and its licensors retain all ownership
 * and intellectual property rights in the accompanying software and in all derivatives hereto.
 *
 * This software and any accompanying information is for suggestion only.  It does not modify Microchip?s standard
 * warranty for its products.  You agree that you are solely responsible for testing the software and determining its
 * suitability.  Microchip has no obligation to modify, test, certify, or support the software.
 *
 * THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * BUT NOT LIMITED TO, IMPLIED WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A PARTICULAR PURPOSE
 * APPLY TO THIS SOFTWARE, ITS INTERACTION WITH MICROCHIP?S PRODUCTS, COMBINATION WITH ANY OTHER PRODUCTS, OR USE IN
 * ANY APPLICATION.
 *
 * IN NO EVENT, WILL MICROCHIP BE LIABLE, WHETHER IN CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE OR BREACH OF
 * STATUTORY DUTY), STRICT LIABILITY, INDEMNITY, CONTRIBUTION, OR OTHERWISE, FOR ANY INDIRECT, SPECIAL, PUNITIVE,
 * EXEMPLARY, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, FOR COST OR EXPENSE OF ANY KIND WHATSOEVER RELATED TO THE
 * SOFTWARE, HOWSOEVER CAUSED, EVEN IF MICROCHIP HAS BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE FORESEEABLE.
 * TO THE FULLEST EXTENT ALLOWABLE BY LAW, MICROCHIP'S TOTAL LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO THIS
 * SOFTWARE WILL NOT EXCEED THE AMOUNT OF FEES, IF ANY, THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR THIS SOFTWARE.
 *
 * MICROCHIP PROVIDES THIS SOFTWARE CONDITIONALLY UPON YOUR ACCEPTANCE OF THESE TERMS.
 *
 *
 * File:        LCD.c
 * Date:        July 22, 2013
 * Compiler:    XC16 v1.10
 *
 * LCD functions for Explorer-16 board
 *
 */

#include <xc.h>
#include "definitions.h"
#include "lcd.h"
#include "timers.h"

// Control signal data pins 
#define  LCD_RW         LATDbits.LATD5       // LCD R/W signal
#define  LCD_RS         LATBbits.LATB15      // LCD RS signal
#define  LCD_EN         LATDbits.LATD4       // LCD E signal

// Control signal pin direction 
#define  LCD_RW_TRIS    TRISDbits.TRISD5
#define  LCD_RS_TRIS    TRISBbits.TRISB15
#define  LCD_EN_TRIS    TRISDbits.TRISD4

// Data signals and pin direction
#define  LCD_DATA_BUS   LATE                // Port for LCD data
#define  LCD_DATA_TRIS  TRISE               // I/O setup for data Port

#define LCD_MAX_COLUMN      16

#define LCD_COMMAND_CLEAR_SCREEN        0x01
#define LCD_COMMAND_RETURN_HOME         0x02
#define LCD_COMMAND_ENTER_DATA_MODE     0x06
#define LCD_COMMAND_CURSOR_OFF          0x0C
#define LCD_COMMAND_CURSOR_ON           0x0F
#define LCD_COMMAND_MOVE_CURSOR_LEFT    0x10
#define LCD_COMMAND_MOVE_CURSOR_RIGHT   0x14
#define LCD_COMMAND_SET_MODE_8_BIT      0x38
#define LCD_COMMAND_ROW_0_HOME          0x80
#define LCD_COMMAND_ROW_1_HOME          0xC0

/* Private Functions *************************************************/
static void LCD_CarriageReturn(void) ;
static void LCD_ShiftCursorLeft(void) ;
static void LCD_ShiftCursorRight(void) ;
static void LCD_ShiftCursorUp(void) ;
static void LCD_ShiftCursorDown(void) ;
static void LCD_SendCommand(char);
static void LCD_SendData(char);

/* Private variables ************************************************/
static uint8_t row, column;

//**********************************************************************************************************************
// Initialize the LCD

void LCD_Initialize(void)
{
    // 30mS delay after Vdd reaches nnVdc before proceeding with LCD initialization
    // not always required and is based on system Vdd rise rate
    timersWait1ms(30);                                                          // 30ms delay

    /* set initial states for the data and control pins */
    LCD_DATA_BUS &= 0xFF00;
    LCD_RW = 0;                                                                 // R/W state set low
    LCD_RS = 0;                                                                 // RS state set low
    LCD_EN = 0;                                                                 // E state set low

    /* set data and control pins to outputs */
    LCD_DATA_TRIS &= 0xFF00;
    LCD_RW_TRIS = 0;                                                            // RW pin set as output
    LCD_RS_TRIS = 0;                                                            // RS pin set as output
    LCD_EN_TRIS = 0;                                                            // E pin set as output

    LCD_SendCommand(LCD_COMMAND_SET_MODE_8_BIT);                                // Set function 2 line 5x7 dots
    LCD_SendCommand(LCD_COMMAND_SET_MODE_8_BIT);                                // Repeat for some unknown reason!
    LCD_SendCommand(LCD_COMMAND_CURSOR_OFF);                                    // Display on/off control, cursor blink off
    LCD_SendCommand(LCD_COMMAND_CLEAR_SCREEN);                                  // Clear display
    LCD_SendCommand(LCD_COMMAND_ENTER_DATA_MODE);                               // Set entry mode increment on shift off
}

//**********************************************************************************************************************
// Write a single command byte to the LCD

static void LCD_SendCommand(char command)
{
    //	TRISD &= 0xFF00;                                                        // ensure RD0 - RD7 are outputs
    LCD_RS = 0;                                                                 // RS is 0 for control register
    LCD_RW = 0;                                                                 // RW is 0 for write
    LCD_DATA_BUS &= 0xFF00;
    LCD_DATA_BUS |= command;                                                    // Put command byte on the data bus
    LCD_EN = 1;                                                                 // Enable pin high
    Nop();                                                                      // Short delay for enable pulse
    Nop();
    Nop();
    Nop();
    Nop();
    LCD_EN = 0;                                                                 // Enable pin low
    timersWait1ms(5);                                                           // 5ms delay
}

//**********************************************************************************************************************
// Write a single data byte to the LCD

static void LCD_SendData(char data)
{
    //	TRISD &= 0xFF00;                                                        // ensure RD0 - RD7 are outputs
    LCD_RS = 1;                                                                 // RS is 1 for data
    LCD_RW = 0;                                                                 // RW is 0 for write
    LCD_DATA_BUS &= 0xFF00;
    LCD_DATA_BUS |= data;                                                       // Put data byte on the data bus
    LCD_EN = 1;                                                                 // Enable pin high
    Nop();                                                                      // Short delay for enable pulse
    Nop();
    Nop();
    Nop();
    Nop();
    LCD_EN = 0;                                                                 // Enable pin low
    LCD_RS = 0;                                                                 // negate register select to 0
    timersWait1us(400);                                                         // 400uS delay
}

//**********************************************************************************************************************
// Write a single data byte to the LCD and handle <CR> <LF> and <BS> characters

void LCD_PutChar ( char inputCharacter )
{
    switch (inputCharacter)
    {
        case '\r':
            LCD_CarriageReturn ( ) ;
            break ;

        case '\n':
            if (row == 0)
            {
                LCD_ShiftCursorDown ( ) ;
            }
            else
            {
                LCD_ShiftCursorUp ( ) ;
            }
            break ;

        case '\b':
            LCD_ShiftCursorLeft ( ) ;
            LCD_PutChar ( ' ' ) ;
            LCD_ShiftCursorLeft ( ) ;
            break ;

        default:
            LCD_SendData ( inputCharacter ) ;
            column++ ;

            if (column == LCD_MAX_COLUMN)
            {
                column = 0 ;
                if (row == 0)
                {
                    LCD_SendCommand (LCD_COMMAND_ROW_1_HOME) ;
                    row = 1 ;
                }
                else
                {
                    LCD_SendCommand (LCD_COMMAND_ROW_0_HOME) ;
                    row = 0 ;
                }
            }
            break ;
    }
}

//**********************************************************************************************************************
// Write a string of text to the LCD

void LCD_WriteString(char *data) {
    uint8_t i;
    for (i = 0; i < 18; i++) {                                                  //Write up to 16 bytes (16 chars plus <CR> <LF>)
        if (*data != '\0')                                                      //Check for the end of the text string
            LCD_PutChar(*data++);                                               //Write the byte if it is not null
        else                                                                    //else
            break;                                                              //we are done if the byte is null
    }
}

static void LCD_CarriageReturn ( void )
{
    if (row == 0) {
        LCD_SendCommand(LCD_COMMAND_ROW_0_HOME);
    }
    else {
        LCD_SendCommand(LCD_COMMAND_ROW_1_HOME);
    }
    column = 0 ;
}

static void LCD_ShiftCursorLeft ( void )
{
    uint8_t i ;

    if (column == 0)
    {
        if (row == 0)
        {
            LCD_SendCommand ( LCD_COMMAND_ROW_1_HOME) ;
            row = 1 ;
        }
        else
        {
            LCD_SendCommand ( LCD_COMMAND_ROW_0_HOME) ;
            row = 0 ;
        }

        //Now shift to the end of the row
        for (i = 0 ; i < ( LCD_MAX_COLUMN - 1 ) ; i++)
        {
            LCD_ShiftCursorRight ( ) ;
        }
    }
    else
    {
        column-- ;
        LCD_SendCommand ( LCD_COMMAND_MOVE_CURSOR_LEFT) ;
    }
}

static void LCD_ShiftCursorRight ( void )
{
    LCD_SendCommand ( LCD_COMMAND_MOVE_CURSOR_RIGHT) ;
    column++ ;

    if (column == LCD_MAX_COLUMN) {
        column = 0 ;
        if (row == 0) {
            LCD_SendCommand ( LCD_COMMAND_ROW_1_HOME) ;
            row = 1 ;
        }
        else {
            LCD_SendCommand ( LCD_COMMAND_ROW_0_HOME) ;
            row = 0 ;
        }
    }
}

static void LCD_ShiftCursorUp ( void )
{
    row = 0;
    LCD_SendCommand ( LCD_COMMAND_ROW_0_HOME + column);
}

static void LCD_ShiftCursorDown ( void )
{
    row = 1;
    LCD_SendCommand ( LCD_COMMAND_ROW_1_HOME + column);
}

void LCD_ClearScreen(void)
{
    LCD_SendCommand(LCD_COMMAND_CLEAR_SCREEN);
    LCD_SendCommand(LCD_COMMAND_RETURN_HOME);
    LCD_SendCommand(LCD_COMMAND_ROW_0_HOME);
    row = 0;
    column = 0;
}

