/******************************************************************************
Software License Agreement

The software supplied herewith by Microchip Technology Incorporated
(the "Company") for its PIC(R) Microcontroller is intended and
supplied to you, the Company's customer, for use solely and
exclusively on Microchip PICmicro Microcontroller products. The
software is owned by the Company and/or its supplier, and is
protected under applicable copyright laws. All rights are reserved.
Any use in violation of the foregoing restrictions may subject the
user to criminal sanctions under applicable laws, as well as to
civil liability for the breach of the terms and conditions of this
license.

THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION. NO WARRANTIES,
WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
*******************************************************************************/

#include <xc.h>
#include "definitions.h"
#include "lcd.h"
#include "timers.h"

/* Private Definitions ***********************************************/
#define LCD_MAX_COLUMN      16

#define LCD_COMMAND_CLEAR_SCREEN        0x01
#define LCD_COMMAND_RETURN_HOME         0x02
#define LCD_COMMAND_ENTER_DATA_MODE     0x06
#define LCD_COMMAND_CURSOR_OFF          0x0C
#define LCD_COMMAND_CURSOR_ON           0x0F
#define LCD_COMMAND_MOVE_CURSOR_LEFT    0x10
#define LCD_COMMAND_MOVE_CURSOR_RIGHT   0x14
#define LCD_COMMAND_SET_MODE_8_BIT      0x38
#define LCD_COMMAND_ROW_0_HOME          0x80
#define LCD_COMMAND_ROW_1_HOME          0xC0

/* Private Functions *************************************************/
static void LCD_CarriageReturn ( void ) ;
static void LCD_ShiftCursorLeft ( void ) ;
static void LCD_ShiftCursorRight ( void ) ;
static void LCD_ShiftCursorUp ( void ) ;
static void LCD_ShiftCursorDown ( void ) ;
static void LCD_SendCommand(char);
static void LCD_SendData(char);

/* Private variables ************************************************/
static uint8_t row, column;

//**********************************************************************************************************************
// Initialize the LCD

void LCD_Initialize ( void )
{
    PMMODE = 0x03ff ;                                                           //PMP in master mode, max wait states
    PMCON = 0x8383 ;                                                            //Enable PMP, no addr mux, enable RdWr port, enable Enb port, no CS, RdWr and Enb active high
    PMAEN = 0x0001 ;                                                            //Enable A0

    timersWait1ms(60);                                                          //60ms delay for LCD to power up 
    LCD_SendCommand(LCD_COMMAND_SET_MODE_8_BIT);
    timersWait1ms(30);                                                          //30ms delay after setting mode 
    LCD_SendCommand(LCD_COMMAND_CURSOR_OFF);
    LCD_SendCommand(LCD_COMMAND_ENTER_DATA_MODE);
    LCD_ClearScreen();
}

//**********************************************************************************************************************
// Write a string of text to the LCD

void LCD_WriteString(char *data) {
    uint8_t i;
    for (i = 0; i < 18; i++) {                                                  //Write up to 16 bytes (16 chars plus <CR> <LF>)
        if (*data != '\0') {                                                    //Check for the end of the text string
            LCD_PutChar(*data++);                                               //Write the byte if it is not null
        }
        else {                                                                  //else
            break;                                                              //we are done if the byte is null
        }
    }
}

//**********************************************************************************************************************
// Write a single data byte to the LCD and handle <CR> <LF> and <BS> characters

void LCD_PutChar ( char inputCharacter )
{
    switch (inputCharacter)
    {
        case '\r':
            LCD_CarriageReturn();
            break;

        case '\n':
            if (row == 0) {
                LCD_ShiftCursorDown();
            }
            else {
                LCD_ShiftCursorUp();
            }
            break;

        case '\b':
            LCD_ShiftCursorLeft();
            LCD_PutChar(' ');
            LCD_ShiftCursorLeft();
            break;

        default:
            LCD_SendData(inputCharacter);
            column++;

            if (column == LCD_MAX_COLUMN) {
                column = 0;
                if (row == 0) {
                    LCD_SendCommand(LCD_COMMAND_ROW_1_HOME);
                    row = 1;
                }
                else {
                    LCD_SendCommand(LCD_COMMAND_ROW_0_HOME);
                    row = 0;
                }
            }
            break;
    }
}

//**********************************************************************************************************************
// Clears the screen

void LCD_ClearScreen ( void )
{
    LCD_SendCommand(LCD_COMMAND_CLEAR_SCREEN);
    LCD_SendCommand(LCD_COMMAND_RETURN_HOME);
    row = 0 ;
    column = 0 ;
}

//**********************************************************************************************************************
// Handles a carriage return

 static void LCD_CarriageReturn ( void )
{
    if (row == 0) {
        LCD_SendCommand(LCD_COMMAND_ROW_0_HOME);
    }
    else {
        LCD_SendCommand(LCD_COMMAND_ROW_1_HOME);
    }
    column = 0 ;
}

//**********************************************************************************************************************
// Shifts cursor left one spot (wrapping if required)
static void LCD_ShiftCursorLeft ( void )
{
    uint8_t i ;

    if (column == 0) {
        if (row == 0) {
            LCD_SendCommand(LCD_COMMAND_ROW_1_HOME);
            row = 1;
        }
        else {
            LCD_SendCommand(LCD_COMMAND_ROW_0_HOME);
            row = 0;
        }
        //Now shift to the end of the row
        for (i = 0; i < ( LCD_MAX_COLUMN - 1 ); i++) {
            LCD_ShiftCursorRight();
        }
    }
    else {
        column--;
        LCD_SendCommand(LCD_COMMAND_MOVE_CURSOR_LEFT);
    }
}

//**********************************************************************************************************************
// Shifts cursor right one spot (wrapping if required)

static void LCD_ShiftCursorRight(void)
{
    LCD_SendCommand(LCD_COMMAND_MOVE_CURSOR_RIGHT);
    column++;

    if (column == LCD_MAX_COLUMN) {
        column = 0;
        if (row == 0) {
            LCD_SendCommand(LCD_COMMAND_ROW_1_HOME);
            row = 1;
        }
        else {
            LCD_SendCommand(LCD_COMMAND_ROW_0_HOME);
            row = 0;
        }
    }
}

//**********************************************************************************************************************
// Shifts cursor up one spot (wrapping if required)

static void LCD_ShiftCursorUp(void)
{
    uint8_t i;

    for (i = 0; i < LCD_MAX_COLUMN; i++) {
        LCD_ShiftCursorLeft();
    }
}
//**********************************************************************************************************************
// Shifts cursor down one spot (wrapping if required)

static void LCD_ShiftCursorDown(void)
{
    uint8_t i;

    for (i = 0; i < LCD_MAX_COLUMN; i++) {
        LCD_ShiftCursorRight();
    }
}

//**********************************************************************************************************************
// Write a single command byte to the LCD

static void LCD_SendCommand(char command)
{
    PMADDR = 0x0000;
    PMDIN1 = command;
    timersWait1ms(2);                                                           //2ms delay
}

//**********************************************************************************************************************
// Write a single data byte to the LCD

static void LCD_SendData(char data)
{
    PMADDR = 0x0001;
    PMDIN1 = data;
    timersWait1ms(2);
//    timersWait1us(400);                                                         // 400uS delay
}

