/*
 * MICROCHIP SOFTWARE NOTICE AND DISCLAIMER:  You may use this software, and any derivatives created by any person or
 * entity by or on your behalf, exclusively with Microchip?s products.  Microchip and its licensors retain all ownership
 * and intellectual property rights in the accompanying software and in all derivatives hereto.
 *
 * This software and any accompanying information is for suggestion only.  It does not modify Microchip?s standard
 * warranty for its products.  You agree that you are solely responsible for testing the software and determining its
 * suitability.  Microchip has no obligation to modify, test, certify, or support the software.
 *
 * THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * BUT NOT LIMITED TO, IMPLIED WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A PARTICULAR PURPOSE
 * APPLY TO THIS SOFTWARE, ITS INTERACTION WITH MICROCHIP?S PRODUCTS, COMBINATION WITH ANY OTHER PRODUCTS, OR USE IN
 * ANY APPLICATION.
 *
 * IN NO EVENT, WILL MICROCHIP BE LIABLE, WHETHER IN CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE OR BREACH OF
 * STATUTORY DUTY), STRICT LIABILITY, INDEMNITY, CONTRIBUTION, OR OTHERWISE, FOR ANY INDIRECT, SPECIAL, PUNITIVE,
 * EXEMPLARY, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, FOR COST OR EXPENSE OF ANY KIND WHATSOEVER RELATED TO THE
 * SOFTWARE, HOWSOEVER CAUSED, EVEN IF MICROCHIP HAS BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE FORESEEABLE.
 * TO THE FULLEST EXTENT ALLOWABLE BY LAW, MICROCHIP'S TOTAL LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO THIS
 * SOFTWARE WILL NOT EXCEED THE AMOUNT OF FEES, IF ANY, THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR THIS SOFTWARE.
 *
 * MICROCHIP PROVIDES THIS SOFTWARE CONDITIONALLY UPON YOUR ACCEPTANCE OF THESE TERMS.
 *
 *
 * File:        main.c
 * Date:        March 17, 2016
 * Compiler:    XC16 v1.26
 *
 * Main file for BM70 Bluetooth LE PICtail Stopwatch application on an Explorer-16 board.
 * 
 * All UART communications with the BM70 is also sent out the RS-232 DB9 connector at 115200 baud.
 * Tested with BM70 firmware v1.04
 *
 */

#if __XC16_VERSION__ != 1026
#error "Incorrect compiler version. Code tested with XC16 v1.26"
#endif
#ifndef __PIC24FJ128GA010__
#error "Incorrect processor. Code written for PIC24FJ128GA010"
#endif

#include <xc.h>
#include "definitions.h"
#include "config bits.h"
#include "timers.h"
#include "uart.h"
#include "bm70.h"
#include "lcd.h"
#include "print.h"
#include "switches.h"
#include "stdlib.h"

void mainInitialize(void);                                                      //Initialize function is below main() in this file
void mainSendStopwatch(uint8_t, uint8_t, uint32_t);

//**********************************************************************************************************************
// Main routine - start of executable code

int main(void) {

    struct MessageIn messageIncoming;                                           //Structure to hold incoming message
    struct ConnectionParameters {                                               //Structure to hold parameters associated with a connection
        uint8_t connectionHandle;                                               //Connection handle
        uint8_t role;                                                           //Master 0 or slave 1
        uint8_t addressType;                                                    //Public 0, random 1, or paired 2
        uint8_t address[6];                                                     //Address of peer (connected) device
        uint16_t connectionInterval;                                            //Connection interval
        uint16_t connectionLatency;                                             //Connection latency
        uint16_t supervisionTimeout;                                            //Supervision timeout
    } connectionParameters;
    
    enum SetAdvertisingEnable enterStandby = EnterStandbyMode;
    enum StateOperating {StateUnconnected, StateWaitForConnection, StateConnected, StatePowerDown} stateOperating = StateUnconnected;
    enum StateStopwatch {StateReset, StateRunning, StateLapTime, StateStopped} stateStopwatch = StateReset;
    uint16_t i = 0;                                                             //General purpose index and counter
    uint8_t btName[7], btAddress[6];                                            //Module initialization returns name and MAC address
    int8_t timerMain;                                                           //Timer variable
    int8_t displayUpdate = TRUE, remoteStart = FALSE;                           //Need to update the stopwatch display, Start stopwatch over Bluetooth
    uint32_t stopwatchTime = 0;
    
    mainInitialize();                                                           //Initialize the pins and peripherals
    LCD_Initialize();                                                           //Initialize the LCD display
    printText("Initializing...");

    if(bmSetupModule(btName, btAddress) == TRUE) {                              //Set up the BT module and see if it completed successfully
        printDeviceNameAndAddress(btName, btAddress);                           //Show device name and MAC address
    }
    else {                                                                      //Failed to set up the BT module so show message
        printText("Initialize Error");
        SYS_DEBUG_BreakPoint();
        while(1);                                                               //Wait for hardware reset
    }

    timerMain = timersGetAndStartTimer(50);                                     //Get a timer to use in main routine and set it for 50ms
    
    bmSendCommand(SetAdvertisingEnable, (uint8_t)enterStandby, 0);                  //Send command to enter standby mode and start advertising
    //stateOperating = StateWaitForConnection;
    timersResetStopwatch();                                                     //Reset the stopwatch timer

    //------------------------------------------------------------------------------------------------------------------
    // Main loop

    while (1) {
        bmIncomingTasks();                                                      //Process incoming packets from the module
        bmOutgoingTasks();                                                      //Process outgoing packets to the module
        switchTasks();                                                          //Sample the switches

        if (stateOperating == StateConnected || stateOperating == StateUnconnected) { //Check if we are in a valid state for stopwatch operation
            if (remoteStart == TRUE) {                                          //See if Bluetooth request to start timer
                remoteStart = FALSE;
                if (stateStopwatch != StateReset) {                             //See if timer is not already reset 
                    timersResetStopwatch();                                     //Reset the stopwatch
                    stopwatchTime = 0;                                          //Stopwatch time is zero
                }
                timersStartStopwatch();                                         //Start the stopwatch
                if (stateOperating == StateConnected) {
                    mainSendStopwatch(connectionParameters.connectionHandle, 1, stopwatchTime);
                }
                stateStopwatch = StateRunning;                                  //Now in running state
            }
            if (switchS3WasPressed() == TRUE) {                                 //See start/stop switch was pressed
                switch (stateStopwatch) {
                    case StateReset:                                            //In reset state
                        timersStartStopwatch();                                 //so start the stopwatch
                        if (stateOperating == StateConnected) {
                            mainSendStopwatch(connectionParameters.connectionHandle, 1, stopwatchTime);
                        }
                        stateStopwatch = StateRunning;                          //Now in running state
                        break;
                        
                    case StateRunning:                                          //In running state
                    case StateLapTime:                                          //or in lap time state
                        stopwatchTime = timersStopStopwatch();                  //so stop the stopwatch and record the time
                        if (stateOperating == StateConnected) {
                            mainSendStopwatch(connectionParameters.connectionHandle, 2, stopwatchTime); //Stop
                        }
                        displayUpdate = TRUE;                                   //Not running but need to update the display once
                        stateStopwatch = StateStopped;                          //Now in stopped state
                        break;
                        
                    case StateStopped:                                          //In stopped state
                        timersStartStopwatch();                                 //so start the stopwatch
                        if (stateOperating == StateConnected) {
                            mainSendStopwatch(connectionParameters.connectionHandle, 1, stopwatchTime); //Start
                        }
                        stateStopwatch = StateRunning;                          //Now in running state
                        break;

                    default:                                                    //Invalid state
                        timersResetStopwatch();                                 //so reset the stopwatch
                        stopwatchTime = 0;                                      //Stopwatch time is zero
                        if (stateOperating == StateConnected) {
                            mainSendStopwatch(connectionParameters.connectionHandle, 4, stopwatchTime); //Reset
                        }
                        displayUpdate = TRUE;                                   //Not running but need to update the display once
                        stateStopwatch = StateReset;                            //Now in reset state
                        break;
                }
            }
            if (switchS6WasPressed() == TRUE) {                                 //See reset/lap switch was pressed
                switch (stateStopwatch) {
                    case StateReset:                                            //In reset state
                        break;                                                  //so do not respond to switch press
                        
                    case StateRunning:                                          //In running state
                    case StateLapTime:                                          //or in lap time state
                        stopwatchTime = timersReadStopwatch();                  //so read the stopwatch and record the time
                        if (stateOperating == StateConnected) {
                            mainSendStopwatch(connectionParameters.connectionHandle, 3, stopwatchTime); //Lap time
                        }
                        displayUpdate = TRUE;                                   //Not running but need to update the display once
                        stateStopwatch = StateLapTime;                          //Now in lap time state
                        break;
                        
                    case StateStopped:                                          //In stopped state
                    default:                                                    //or invalid state
                        timersResetStopwatch();                                 //so reset the stopwatch
                        stopwatchTime = 0;                                      //Stopwatch time is zero
                        if (stateOperating == StateConnected) {
                            mainSendStopwatch(connectionParameters.connectionHandle, 4, stopwatchTime); //Reset
                        }
                        displayUpdate = TRUE;                                   //Not running but need to update the display once
                        stateStopwatch = StateReset;                            //Now in reset state
                        break;
                }
            }
        }

        if (displayUpdate == TRUE) {                                            //Stopwatch display is not incrementing but need to update the display once
            displayUpdate = FALSE;                                              
            printStopwatch(stopwatchTime, FALSE);                               //Update the stopwatch time on the display
        }
        
        if (stateStopwatch == StateRunning) {                                   //Stopwatch is running so need to update the display periodically
            if (timersIsTimerDone(timerMain)) {                                 //Update LCD stopwatch every 210ms to roll through all lower digits
                timersStartTimer(timerMain, 210);                               //Use software timer to set next sample delay
                stopwatchTime = timersReadStopwatch();                          //Read the stopwatch
                printStopwatch(stopwatchTime, FALSE);                           //Update the stopwatch time on the display
            }
        }
        else if (stateStopwatch == StateLapTime) {                              //Stopwatch is timing laps so need to update the running indicator on the display periodically
            if (timersIsTimerDone(timerMain)) {                                 //Update LCD stopwatch every 500ms to flash the indicator
                timersStartTimer(timerMain, 500);                               //Use software timer to set next sample delay
                printStopwatch(stopwatchTime, TRUE);                            //Reprint the stopwatch time on the display with the running indicator
            }
        }
        
        if (bmGetIncomingMessage(&messageIncoming) == TRUE) {                   //See if we have a message from the module
            switch (messageIncoming.event) {
                case NoEventOpcode:
                    SYS_DEBUG_BreakPoint();
                case PasskeyEntryRequest:
                case PairingComplete:
                case PasskeyConfirmRequest:
                case AdvertisingReport:
                    break;
                case LeConnectionComplete:
                    if (messageIncoming.data[0] != 0 || messageIncoming.length < 16) {
                        printText(" Connect Failed");
                        break;
                    }
                    printText(" Connected");
                    connectionParameters.connectionHandle = messageIncoming.data[1];
                    connectionParameters.role = messageIncoming.data[2];
                    connectionParameters.addressType = messageIncoming.data[3];
                    for (i = 0; i < 6; i++) {
                        connectionParameters.address[i] = messageIncoming.data[i + 4];
                    }        
                    connectionParameters.connectionInterval = messageIncoming.data[10] * 256 + messageIncoming.data[11];
                    connectionParameters.connectionLatency = messageIncoming.data[12] * 256 + messageIncoming.data[13];
                    connectionParameters.supervisionTimeout = messageIncoming.data[14] * 256 + messageIncoming.data[15];
                    stateOperating = StateConnected;
                    break;
                case DisconnectionComplete:
                    printText(" Disconnected");
                    bmSendCommand(SetAdvertisingEnable, (uint8_t)enterStandby, 0);  //Send command to enter standby mode and start advertising
                    //stateOperating = StateWaitForConnection;
                    stateOperating = StateUnconnected;
                    break;
                case ConnectionParameterUpdateNotify:
                case CommandComplete:
                    break;
                case StatusReport:
                    switch (messageIncoming.data[0]) {
                        case NoStatusResponse:
                        case ScanningMode:
                        case ConnectingMode:
                        case StandbyMode:
                        case BroadcastMode:
                        case TransparentServiceEnabledMode:
                        case IdleMode:
                        case ShutdownMode:
                        case ConfigureMode:
                        case BleConnectedMode:
                        default:
                            break;
                    }
                    break;
                case ConfigureModeStatus:
                case DiscoverAllPrimaryServiceResponse:
                case DiscoverSpecificPrimaryServiceCharacteristicResponse:
                case DiscoverAllCharacteristicDescriptorResponse:
                case CharacteristicValueReceived:
                    break;
                case ClientWriteCharacteristicValue:
                    if (messageIncoming.data[1] == (PRIVATE_CHAR_IN_VAL_HANDLE >> 8) && messageIncoming.data[2] == (PRIVATE_CHAR_IN_VAL_HANDLE & 0x00ff)) { //See if handle is 0x8002 for our writable characteristic
                        if (messageIncoming.data[3] == 0x01) {
                            remoteStart = TRUE;
                        }
                    }
                    break;
                case ReceivedTransparentData:
                default:
                    printTextAndNumber("Event: ", messageIncoming.event);       //Unexpected event opcode
                    break;
            }
        }
        
        if (BLE_NOT_CONNECT_PIN == 0) {                                         //See if there is a Bluetooth connection
            LED_D3 = 1;                                                         //Light a LED to indicate connected
        }
        else {
            LED_D3 = 0;                                                         //Turn off LED when not connected
        }
    } //End of while(1) main loop
    return (TRUE);
}


//**********************************************************************************************************************
// Initialize the pins and peripherals

void mainInitialize(void)
{
    AD1PCFG = 0xffcf;                                                           //All digital I/O except AN4 for temperature sensor and AN5 for potentiometer
    LATA = 0x0000;                                                              //All LEDs off
    TRISA = 0xff00;                                                             //RA0-RA7 are active high LED outputs
    LATF = 0x01e8;                                                              //RF3 U1TX high, RF5 U2TX high, RF6 BM70 PAIRING_KEY high, RF7 BM70 RX_IND high, RF8 LINK_DROP high
    TRISF = 0xfe17;                                                             //RF2 U1RX input
#ifdef USE_BM70
    LATB = 0x0002;                                                              //RB1 Mode high (application mode), RB15 LCD RS low
    TRISB = 0x7ffd;                                                             //RB1, RB15 is an output
    LATD = 0x8000;                                                              //Start with RD15 U1RTS high, RD4 LCD E low, RD5 LCD R/W low
    TRISD = 0x7fff;                                                             //RD15 is U1RTS output, RD14 is U1CTS input
    LATE = 0x0000;                                                              //LCD data pins on RE0-7, BLE_TX_IND_PIN input on RE9
    TRISE = 0xff00;                                                             //LCD data pins are outputs
    LATG = 0x0000;                                                              //RG1 reset pin starts active low
    TRISG = 0xfffd;                                                             //RG1 is active low BM70 reset signal
#else
    LATB = 0x0000;                                                              //RB15 LCD RS low
    TRISB = 0x7fff;                                                             //RB15 is an output
    LATD = 0x0000;                                                              //Start with RD4 LCD E low, RD5 LCD R/W low
    TRISD = 0xffff;                                                             //RD15 is U1RTS output, RD14 is U1CTS input
    LATE = 0x0100;                                                              //LCD data pins on RE0-7,RE8 U1RTS high, BLE_TX_IND_PIN input on RE9
    TRISE = 0xfe00;                                                             //LCD data pins and U1RTS are outputs
    LATG = 0x0004;                                                              //RG1 reset pin starts active low, RB2 Mode high (application mode)
    TRISG = 0xfff9;                                                             //RG1 is active low BM70 reset signal, RB2 Mode signal
#endif
    
    uartInitialize();                                                           //Initialize the UART
    timersInitialize();                                                         //Initialize the timers
    switchInitialize();                                                         //Initialize the sampling of the switches 
    
    INTCON1 = 0;                                                                //Interrupt nesting enabled
    INTCON2 = 0;                                                                //GIE off, ext interrupts on rising edge
}

//**********************************************************************************************************************
// Send the stopwatch state and time

void mainSendStopwatch(uint8_t handle, uint8_t state, uint32_t time)
{
    uint8_t characteristicPacket[8];                                            //Characteristic packet includes connection handle, char value handle, and char value data

    characteristicPacket[0] = handle;                                           //Connection handle
    characteristicPacket[1] = PRIVATE_CHAR_OUT_VAL_HANDLE >> 8;                 //Characteristic value handle is 0x8004
    characteristicPacket[2] = PRIVATE_CHAR_OUT_VAL_HANDLE & 0x00ff;
    characteristicPacket[3] = state;                                            //Start, stop, lap, reset
    characteristicPacket[4] = (uint8_t)((time & 0xff000000) >> 24);             //Stopwatch time
    characteristicPacket[5] = (uint8_t)((time & 0x00ff0000) >> 16);
    characteristicPacket[6] = (uint8_t)((time & 0x0000ff00) >> 8);
    characteristicPacket[7] = (uint8_t)(time & 0x000000ff);
    
    bmSendCommand(SendCharacteristicValue, 5, &characteristicPacket[0]);        //Send command to send characteristic notification over BLE
}


