/*
 * Copyright (C) 2014 Microchip Technology Inc. and its subsidiaries.  You may use this software and any derivatives
 * exclusively with Microchip products.
 *
 * THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS
 * SOFTWARE, INCLUDING ANY IMPLIED WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ITS INTERACTION WITH MICROCHIP PRODUCTS, COMBINATION WITH ANY OTHER PRODUCTS, OR USE IN ANY APPLICATION.
 *
 * IN NO EVENT WILL MICROCHIP BE LIABLE FOR ANY INDIRECT, SPECIAL, PUNITIVE, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE,
 * COST OR EXPENSE OF ANY KIND WHATSOEVER RELATED TO THE SOFTWARE, HOWEVER CAUSED, EVEN IF MICROCHIP HAS BEEN ADVISED OF
 * THE POSSIBILITY OR THE DAMAGES ARE FORESEEABLE.  TO THE FULLEST EXTENT ALLOWED BY LAW, MICROCHIP'S TOTAL LIABILITY ON
 * ALL CLAIMS IN ANY WAY RELATED TO THIS SOFTWARE WILL NOT EXCEED THE AMOUNT OF FEES, IF ANY, THAT YOU HAVE PAID
 * DIRECTLY TO MICROCHIP FOR THIS SOFTWARE.
 *
 * MICROCHIP PROVIDES THIS SOFTWARE CONDITIONALLY UPON YOUR ACCEPTANCE OF THESE TERMS.
 *
 * 
 * File:        timers.c
 * Date:        July 24, 2014
 * Compiler:    XC16 v1.21
 *
 * Timer functions
 * 
 * Timer 1 is used for accurate microsecond interval delays
 * Timer 5 is used to generate several software timers
 *
 */

#include <xc.h>
#include "definitions.h"
#include "timers.h"

#define NUM_TIMERS 4

static volatile uint16_t timerTickCount[NUM_TIMERS] = {0, 0, 0, 0};             //Four general purpose software timers
static volatile uint8_t timerIsInUse[NUM_TIMERS] = {FALSE, FALSE, FALSE, FALSE}; 
static volatile uint32_t timerLong = 0, timerStopwatchCount = 0;
static volatile uint8_t timerStopwatchRunning = FALSE;

//**********************************************************************************************************************
// Initialize the timers

void timersInitialize(void)
{
    int8_t i;
    for (i = 0; i < 4; i++) {
        timerTickCount[i] = 0;
        timerIsInUse[i] = FALSE;
    }
    //Timer 5 is used for interrupt based timers counting 1ms intervals to a resolution of 0.1ms
    T5CON = TIMER_OFF;                                                          //Timer 5 off
    TMR5 = 0;                                                                   //Clear timer 5
    PR5 = TIMER_100US_PERIOD;                                                   //Set the period value for 100us
    T5CON = TIMER_ON_PRESCALE1;                                                 //using 1:1 prescaler and turn on timer 5
    IFS1bits.T5IF = 0;                                                          //Clear the interrupt flag
    IEC1bits.T5IE = 1;                                                          //Enable the timer 5 interrupt (global enable is in main initialization)
}

//**********************************************************************************************************************
// Get one of the general purpose software timers

int8_t timersGetTimer(void)
{
    int8_t timer;
    
    for (timer = 0; timer < NUM_TIMERS; timer++) {                              //Search through each of the software timers
        if (timerIsInUse[timer] == FALSE) {                                     //Check if timer is not in use
            timerTickCount[timer] = 0;                                          //Stop the timer
            timerIsInUse[timer] = TRUE;                                         //Timer is now in use
            return timer;                                                       //Successfully found and started the timer so return timer number
        }
    }
    SYS_DEBUG_BreakPoint();
    LED_PORT = 0x55;                                                            //Diagnostic indication on LEDs
    return ERROR;                                                               //Failed to start timer because they were all in use
}

//**********************************************************************************************************************
// Find and start one of the software timers

int8_t timersGetAndStartTimer(int16_t count)
{
    int8_t timer;
    
    for (timer = 0; timer < NUM_TIMERS; timer++) {                              //Search through each of the software timers
        if (timerIsInUse[timer] == FALSE) {                                     //Check if timer is not in use
            if (count < 6554) {
                timerTickCount[timer] = count * 10;                             //Interrupt is every 100us but StartGeneralTimer() takes multiple of 1ms so multiply by 10;
            }
            else {
                timerTickCount[timer] = 0xffff;                                 //Selected time is beyond max so set to max;
            }
            timerIsInUse[timer] = TRUE;                                         //Timer is now in use
            return timer;                                                       //Successfully found and started the timer so return timer number
        }
    }
    SYS_DEBUG_BreakPoint();
    LED_PORT = 0x55;                                                            //Diagnostic indication on LEDs
    return ERROR;                                                               //Failed to start timer because they were all in use
}

//**********************************************************************************************************************
// Start a general purpose software timer

int8_t timersStartTimer(int8_t timer, int16_t count)
{
    if(timer >= 0 && timer < NUM_TIMERS && timerIsInUse[timer] == TRUE) {       //Check that the timer is valid
        if (count < 6554) {
            timerTickCount[timer] = count * 10;                                 //Interrupt is every 100us but StartGeneralTimer() takes multiple of 1ms so multiply by 10;
        }
        else {
            timerTickCount[timer] = 0xffff;                                     //Selected time is beyond max so set to max;
        }
        return TRUE;                                                            //Successfully ended the timer
    }
    SYS_DEBUG_BreakPoint();
    LED_PORT = 0xaa;                                                            //Diagnostic indication on LEDs
    return ERROR;                                                               //Failed to end timer because it was not a valid timer number
}

//**********************************************************************************************************************
// Free up a general purpose software timer

int8_t timersFreeTimer(int8_t timer)
{
    if(timer >= 0 && timer < NUM_TIMERS && timerIsInUse[timer] == TRUE) {       //Check that the timer is valid
        timerTickCount[timer] = 0;                                              //Stop the timer
        timerIsInUse[timer] = FALSE;                                            //Timer is no longer in use
        return TRUE;                                                            //Successfully freed the timer
    }
    return ERROR;                                                               //Failed to free timer because it was not a valid timer number
}

//**********************************************************************************************************************
// Check if one of the software software timers has timed out

int8_t timersIsTimerDone(int8_t timer)
{
    if(timer >= 0 && timer < NUM_TIMERS && timerIsInUse[timer] == TRUE) {       //Check that the timer is valid
        if(timerTickCount[timer] == 0) {                                        //Check if counted down to zero
            return TRUE;                                                        //Return true
        }
        return FALSE;                                                           //Timer has not timed out so return false
    }
    return ERROR;                                                               //Failed because it was not a valid timer number
}

//**********************************************************************************************************************
// Get the tick count (100us) of one of the software software timers

uint16_t timersGetTimerTickCount(int8_t timer)
{
    if(timer >= 0 && timer < NUM_TIMERS && timerIsInUse[timer] == TRUE) {       //Check that the timer is valid
        return timerTickCount[timer];                                           //Return the tick count
    }
    return 0;                                                                   //or return 0
}


//**********************************************************************************************************************
// Start the long timer

void timersStartLongTimer(int32_t count)
{
    timerLong = count * 10;                                                     //Interrupt is every 100us so multiply by 10;
}

//**********************************************************************************************************************
// Has the long timer timed out

int8_t timersIsLongTimerDone(void)
{
    if(timerLong == 0) {                                                        //Check if counted down to zero
        return TRUE;                                                            //Return true
    }
    return FALSE;                                                               //Timer has not timed out so return false
}


//**********************************************************************************************************************
// Simple delay for n milliseconds

void timersWait1ms(int16_t NumMilliseconds)
{
    int timer;
    
    timer = timersGetAndStartTimer(NumMilliseconds);                            //Find and start software timer
    if(timer != ERROR) {                                                        //CXheck that we got a valid timer
        while(!timersIsTimerDone(timer)) {}                                     //and wait for it to count down
        timersFreeTimer(timer);                                                 //Free up the timer
    }
    else {                                                                      //No timer available so use microsecond timer
        for(timer = 0; timer < NumMilliseconds; timer++) {                      //Repeat 1ms time delay NumMilliseconds times
            timersWait1us(1000);                                                //Wait 1ms
        }
    }
}

//**********************************************************************************************************************
// Simple delay for n microseconds

void timersWait1us(int16_t numMicroseconds)
{
    if(numMicroseconds <= 0)                                                    //Check that the value is valid
        return;                                                                 //else return immediately
    else
    {
        T1CON = TIMER_OFF;                                                      //Use hardware timer 1
        TMR1 = 0;                                                               //Clear the timer
        PR1 = numMicroseconds << 1;                                             //Double to get the count required because we are clocking at 2MHz
        IFS0bits.T1IF = 0;                                                      //Clear the interrupt flag
        T1CON = TIMER_ON_PRESCALE8;                                             //Timer clocks at 2 MHz with 16MHz Fcy
        while(!IFS0bits.T1IF);                                                  //Wait for the interrupt flag to indicate the period has been reached
        T1CON = TIMER_OFF;                                                      //Stop the timer
        IFS0bits.T1IF = 0;                                                      //Clear the interrupt flag
    }
}

//**********************************************************************************************************************
// Reset the stopwatch

void timersResetStopwatch(void)
{
    timerStopwatchRunning = FALSE;                                              //Stop the stopwatch, stop the interrupt from incrementing timerStopwatchCount variable
    timerStopwatchCount = 0;                                                    //Clear the count
}

//**********************************************************************************************************************
// Start the stopwatch

void timersStartStopwatch(void)
{
    timerStopwatchRunning = TRUE;                                               //Allow interrupt service to increment the stopwatch 
}

//**********************************************************************************************************************
// Stop the stopwatch

uint32_t timersStopStopwatch(void)
{
    timerStopwatchRunning = FALSE;                                              //Stop the interrupt from incrementing timerStopwatch variable
    return timerStopwatchCount;                                                 //Return the stopwatch count value
}

//**********************************************************************************************************************
// Read the stopwatch

uint32_t timersReadStopwatch(void)
{
    uint32_t temp = 0;                                                          //Temporary variable because count can get updated asynchronously in the interrupt

    while (temp != timerStopwatchCount) {
        temp = timerStopwatchCount;                                             //Read the stopwatch count
    }
    return temp;
    
}

//**********************************************************************************************************************
// Timer 5 interrupt routine

void _ISR_NO_AUTO_PSV _T5Interrupt(void)
{
    IFS1bits.T5IF = 0;                                                          //Clear the interrupt flag
    if (timerTickCount[0] != 0) {                                               //If general software timer 0 has not reached zero then
        timerTickCount[0]--;                                                    //decrement it
    }
    if (timerTickCount[1] != 0) {                                               //If general software timer 1 has not reached zero then
        timerTickCount[1]--;                                                    //decrement it
    }
    if (timerTickCount[2] != 0) {                                               //If general software timer 2 has not reached zero then
        timerTickCount[2]--;                                                    //decrement it
    }
    if (timerTickCount[3] != 0) {                                               //If general software timer 3 has not reached zero then
        timerTickCount[3]--;                                                    //decrement it
    }
    if (timerLong != 0) {
        timerLong--;
    }
    if (timerStopwatchRunning == TRUE) {
        timerStopwatchCount++;
    }
}

